/**
 * This library handles exclusively the auto reset of the modbus
 * error which calls from the function unSystemIntegrity_unicosPLCErrorCallback
 * in unSystemIntegrity_unicosPLC.ctl
 *
 * The goal is to handle the issue UNSYSTINT-40 by resetting the modbus
 * error automatically. If the reset happens too often in a short period of time,
 * i.e. the error is continuously sent by the PLC, the error is not reset anymore
 * as it seems to be a legitimate error
 *
 * 22-November-2021
 * Jean-Charles Tournier
 * CERN BE-ICS-FT
 **/

// _unSystemIntegrity_autoResetModbusError
/**
 *  Main function called by unSystemIntegrity_unicosPLCErrorCallback
 * string sAlPlc: DP name of the system alarm for the PLC. Only used for debug info.
 * string sDp: internal datapoint of type _Mod_Plc corresponding to the PLC. This is the DP which is reset
 * int state: value of the dpe .ConnState of sDp. Represents the connection status
 * int error: value of the dpe .error of sDp. Represents the current error
 *
 * returns code are mainly used for unit testing
 **/
int _unSystemIntegrity_autoResetModbusError(string sAlPlc, string sDp, int state, int error)
{
	//Clear the error only if the PLC is connected and the error is not already 0
	if ((state == 1 && error != 0) == false)
	{
		return 1;
	}

	//check first that we dont do it too often
	if (_unSystemIntegrity_isTooManyAutoResetError(sAlPlc) == true)
	{
		DebugTN(__FUNCTION__, "[INFO]", "Do not automatically reset modbus error. It has already been reset too many times in a short while for " + sDp, sAlPlc);
		return 2;
	}

	//Do the actual error reset
	DebugTN(__FUNCTION__, "[INFO]", "Resetting manually the PLC error as the connection is re-established for " + sDp + " (" + sAlPlc + ")", "error", error, "state", state);
	if (dpSetWait(sDp + ".Error", 0) != 0)
	{
		DebugTN(__FUNCTION__, "[ERROR]", "Unable to reset PLC error", sDp);
		DebugTN(getLastError());
		return -1;
	}

	return 3;
}

class ModbusErrorCounterClass
{
private
	dyn_time dtReset;

public
	int getNumReset()
	{
		return dynlen(dtReset);
	}

public
	void addNewResetTime()
	{
		dynAppend(dtReset, getCurrentTime());
	}

public
	void resetCounter()
	{
		dynClear(dtReset);
	}

public
	long getTimeDiffFirstLastReset()
	{
		int iNumReset = dynlen(dtReset);
		if (iNumReset <= 1)
		{
			return 999;
		}
		return (long)(period(dtReset[iNumReset]) - period(dtReset[1]));
	}
};

// _unSystemIntegrity_isTooManyAutoResetError
/**
 * Private function used by unSystemIntegrity_unicosPLCErrorCallback
 * to know if the auto error reset can be performed, i.e. it has not been
 * run too many times in a short period of time
 *
 **/
mapping g_unSystemIntegrity_PLC_errorResetCounter; //Mapping with key the PLC name, and value an object of type modbusErrorCounter
bool _unSystemIntegrity_isTooManyAutoResetError(string sPlc)
{
	const int MAX_NUM_RESET = 3; //number of reset allowed in MAX_RESET_INTERVAL
	const long MAX_RESET_INTERVAL = 10;

	ModbusErrorCounterClass mec;
	if (mappingHasKey(g_unSystemIntegrity_PLC_errorResetCounter, sPlc))
	{
		mec = g_unSystemIntegrity_PLC_errorResetCounter[sPlc];
	}

	mec.addNewResetTime();

	bool bIsTooManyReset = true;
	if (mec.getNumReset() < MAX_NUM_RESET)
	{
		bIsTooManyReset = false;
	}
	else if (mec.getTimeDiffFirstLastReset() > MAX_RESET_INTERVAL)
	{
		mec.resetCounter();
		mec.addNewResetTime();
		bIsTooManyReset = false;
	}

	g_unSystemIntegrity_PLC_errorResetCounter[sPlc] = mec;

	return bIsTooManyReset;
}
